/*********************************************************

Name: compress.h
Author: JW
Revised by:
Version: 1.0.003
Date of mod.: 20.07.1998
Status: ready

Description:
Compression utility for TDTable.

Comments:

*********************************************************/
#ifndef _COMPRESS_H_
#define _COMPRESS_H_

#include <fstream.h>
#include <io.h>
#include <fcntl.h>
#include <sys\stat.h>
#include <assert.h>

#define DICT_SIZE 33
#define SHORT_SIZE 11
#define SHORT_QUANT 2
#define LONG_QUANT 6
#define SHORT_FLOAT 6
#define OTHER_VALS -2001020357
#define SPECIAL_VAL -2001020359
#define NULL_VALUE	-2147483639

//********************************************************

/*************************************************

Simple compressor functions:

Compressor(int fhandle,int mode)
			fhandle - file handle; read or write from given position
			mode - 0: write only, 1: read only

void Save(char *)	- saves stream of bits; string contains "011001" symbols
int Load()			- loads one bit from source
void SaveNumber(int)- for storage of integers in either short (SHORT_SIZE bits)
					  or long (32 bits) form (+1 bit)
void SaveQuant(int)	- for storage numbers in range 2-4129 in either short(5 bits)
					  or long (12 bits) form (+1 bit)
void SaveDict(int)	- for storage numbers in range 0-16, variable lenght (1-7 bits)
void SaveVal(int,int) - for storage positive integers on fixed number of bits

int LoadNumber(), LoadQuant()... - analogous for loading values

**************************************************/

class CompressionFailed
{
public:
	CompressionFailed() {}
};

class DecompressionFailed
{
public:
	DecompressionFailed() {}
};

//***********************************************

class Compressor
{
public:
	Compressor(int,int);	// fhandle, mode. mode==0 - write, 1 - read
	Compressor(int*,int,int);	// memory version: buffer,lenght,mode
	~Compressor();

	int Flush();		// returns no. of ints in memory version, else 0
	int Eof();			// returns 1 if no more bits
 
	void Save(char *);	// a string of bits ("0" or "1")
	void Save(int);		// 1 or 0
	int Load();			// one bit
	void SaveNumber(int);	// for short and long values
	int LoadNumber();
	void SaveQuant(int);	// for short and long quantity
	int LoadQuant();
	void SaveDict(int);		// for dictionary entry number
	int LoadDict();
	void SaveDictVal(int);		// for value from dictionary
	int LoadDictVal();
	void SaveVal(int v,int k);	// for natural coding (positive) on k bits (k<32)
	int LoadVal(int k);
	void SaveStr(char *);		// NULL-terminated
	int LoadStr(char *buf,int);	// maximal buffer size
	void SaveFloat(double);			// 64-bit
	double LoadFloat();	
	
	// Values stream compression:
	// first, scan all values and put it one by one into PrepareStream,
	// then do InitSaveStream, then put all values into SaveStream,
	// then do FlushSaveStream.

	void PrepareStream(int);
	void InitSaveStream(int=0);		// parameter=1 -> force no repetition
	void SaveStream(int);
	void FlushSaveStream();
	void InitLoadStream();
	int LoadStream();

private:
	class CompDictionary;
	int cur_bit;
	int mode;
	unsigned int buf;
	int fhnd;

	int rep_flag;
	int repetition,prev_val;
	CompDictionary *dic;

	int min,max,diff_bit,p2,p5,divide,total;	// for stream analysis

	// Memory version

	int *buffer;	
	int b_len;		// size of buffer for compression (in 4-bytes)
	int b_cur_pos;	// current position in buffer

class CompTree
{
public:
	CompTree(int);
	~CompTree();

	int Add(int);

	int value,number;
	CompTree *left,*right;
};


class CompDictionary
{
public:
	CompDictionary();
	~CompDictionary();

	int Add(int);	// returns 1 iff tree>DICT_SIZE
	double CreateDictionary();	// returns % of OTHER_VALS

	int GetRealSize()
	{ return real_dict_size; }
	void SetRealSize(int s)
	{ 
		assert(s<=DICT_SIZE);
		if(s>DICT_SIZE) throw CompressionFailed();
		real_dict_size=s; 
	}

	int vals[DICT_SIZE];

	int Encode(int);

private:
	void Add(CompTree*);
	void AddToDictionary(int,int);

	CompTree *root;

	int nums[DICT_SIZE];
	int real_dict_size;
	int other_vals_index;
	int total;
	int tree_size;
};

};
#endif

/*
Algorithm:

  1. Calculate distributions for each column.
  2. Coding for values in one column:
  <value>={<dict_code>}{<fixed_val>}
  dict_code - dictionary code for this value, see code table,
				supposing dictionary not empty
  fixed_val - real value, supposing dict_code==OTHER_VALS;
				fixed size, scaled.

  3. Coding for whole column:
  <column>=<dictionary_size><dictionary><repetition_flag>
			<minimum><bit_size><power_2><power_5><value_stream><value_stream>...

  dictionary_size - a <r_quantity>-2 value of dictionary size (may be 0)
  <dictionary>=<real_val><real_val>...<real_val>  - fixed (see above) size of table

  <repetition_flag> - 1 if repetition compression is performed, 0 otherwise
  <minimum>=<real_val> - scaling parameter (shift) of other values (see below)
  <bit_size>=<r_quanity> - size in bits of other value codes (see below)
  <power_2> - 2-bit value of scaling factor (see below)
  <power_5> - 2-bit value of scaling factor: factor=2^(power_2)*5^(power_5)
				OTHER_VALS are scaled by: (origrnal_value)=(minimum)+(factor)*(coded_value)

  <value_stream>=<value>						if repetition_flag=0
  <value_stream>=<r_flag>{<r_quantity>}<value>	if repetition_flag=1

  r_flag - 0 if one value, 1 if more (r_quantity is used)

  <r_quantity>=0<short_quantity> | 1<long_quantity>
  short_quantity - a SHORT_QUANT-bit repetition number (+2, i.e. from 2 to 33)
  long_quantity - a LONG_QUANT-bit repetition number (+34 i.e. from 34 to 4129)

  4. File format:
  <identify_section><no_of_objects><no_of_attrs><column><column>...<column>

  identify_section - 4 bytes string: "RSc"+(version number: one byte, starting from "0")
  no_of_objects - <real_val>
  no_of_attrs - <real_val>

  <real_val>=0<short_number> | 1<long_number>

  short_number - a n-bit unsigned value, when n=SHORT_SIZE
  long_number - a 32-bit signed value


*/

